#ifndef STIM_NETWORK_H
#define STIM_NETWORK_H

#include <stdlib.h>
#include <assert.h>
#include <sstream>
#include <fstream>
#include <algorithm>
#include <string.h>
#include <math.h>
#include <stim/math/vec3.h>
#include <stim/visualization/obj.h>
#include <stim/visualization/swc.h>
#include <stim/visualization/cylinder.h>
#include <stim/structures/kdtree.cuh>
#include <stim/cuda/cudatools/timer.h>


#ifdef __CUDACC__
//device gaussian function
__device__ float gaussianFunction(float x, float std){ return expf(-x/(2*std*std));}

//compute metric in parallel
template <typename T>
__global__ void d_metric(T* M, size_t n, T* D, float sigma){
	size_t x = blockDim.x * blockIdx.x + threadIdx.x;
	if(x >= n) return;
	M[x] = 1.0f - gaussianFunction(D[x], sigma);
}

//find the corresponding edge index from array index
__global__ void d_findedge(size_t* I, size_t n, unsigned* R, size_t* E, size_t ne){
	size_t x = blockDim.x * blockIdx.x + threadIdx.x;
	if(x >= n) return;
	unsigned i = 0;
	size_t N = 0;
	for(unsigned e = 0; e < ne; e++){
		N += E[e];
		if(I[x] < N){
			R[x] = i;
			break;
		}
		i++;
	}
}
#endif

//hard-coded factor
int threshold_fac = 0;

namespace stim{
/** This is the a class that interfaces with gl_spider in order to store the currently
 *   segmented network. The following data is stored and can be extracted:
 *   1)Network geometry and centerline.
 *   2)Network connectivity (a graph of nodes and edges), reconstructed using ANN library.
*/

template<typename T>
class network{

	///Each edge is a fiber with two nodes.
	///Each node is an in index to the endpoint of the fiber in the nodes array.
	class edge : public cylinder<T>
	{
		public:
		unsigned v[2];		//unique id's designating the starting and ending
		// default constructor
		edge() : cylinder<T>() {
			v[1] = (unsigned)(-1); v[0] = (unsigned)(-1);
		}
		/// Constructor - creates an edge from a list of points by calling the stim::fiber constructor

		///@param p is an array of positions in space
		edge(centerline p) : cylinder<T>(p){}

		/// Copy constructor creates an edge from a cylinder
		edge(stim::cylinder<T> f) : cylinder<T>(f) {}

		/// Resamples an edge by calling the fiber resampling function
		edge resample(T spacing){
			edge e(cylinder<T>::resample(spacing));	//call the fiber->edge constructor
			e.v[0] = v[0];					//copy the vertex data
			e.v[1] = v[1];

			return e;						//return the new edge
		}

		/// Output the edge information as a string
		std::string str(){
			std::stringstream ss;
			ss<<"("<<cylinder<T>::size()<<")\tl = "<<this->length()<<"\t"<<v[0]<<"----"<<v[1];
			return ss.str();
		}

		std::vector<edge> split(unsigned int idx){
		
			std::vector< stim::cylinder<T> > C;
			C.resize(2);
			C =	(*this).cylinder<T>::split(idx);
			std::vector<edge> E(C.size());

			for(unsigned e = 0; e < E.size(); e++){
				E[e] = C[e];
			}
			return E;
		}

	};

	///Node class that stores the physical position of the node as well as the edges it is connected to (edges that connect to it), As well as any additional data necessary.
	class vertex : public stim::vec3<T>
	{
		public:
			//std::vector<unsigned int> edges;					//indices of edges connected to this node.
			std::vector<unsigned int> e[2];						//indices of edges going out (e[0]) and coming in (e[1])
			//stim::vec3<T> p;							//position of this node in physical space.

			//constructor takes a stim::vec
			vertex(stim::vec3<T> p) : stim::vec3<T>(p){}

			/// Output the vertex information as a string
			std::string	str(){
				std::stringstream ss;
				ss<<"\t(x, y, z) = "<<stim::vec3<T>::str();

				if(e[0].size() > 0){
					ss<<"\t> ";
					for(unsigned int o = 0; o < e[0].size(); o++)
						ss<<e[0][o]<<" ";
				}
				if(e[1].size() > 0){
					ss<<"\t< ";
					for(unsigned int i = 0; i < e[1].size(); i++)
						ss<<e[1][i]<<" ";
				}

				return ss.str();
			}


	};

protected:

	std::vector<edge> E;       //list of edges
	std::vector<vertex> V;	   //list of vertices.
	//std::vector<int> NT;	   //stores a list of neuronal type for each point in the centerline (will set value only when centerline is built from swc file)

public:

	///default constructor
	network()
	{
		
	}

	///constructor with a file to load.
	network(std::string fileLocation)
	{
		load_obj(fileLocation);
	}

	///Returns the number of edges in the network.
	unsigned int edges(){
		return E.size();
	}

	///Returns the number of nodes in the network.
	unsigned int vertices(){
		return V.size();
	}

	//scale the network by some constant value
	//	I don't think these work??????
	/*std::vector<vertex> operator*(T s){
		for (unsigned i=0; i< vertices; i ++ ){
			V[i] = V[i] * s;
		}
		return V;
	}

	std::vector<vertex> operator*(vec<T> s){
		for (unsigned i=0; i< vertices; i ++ ){
			for (unsigned dim = 0 ; dim< 3; dim ++){
				V[i][dim] = V[i][dim] * s[dim];
			}
		}
		return V;
	}*/

	// Returns an average of branching index in the network
	double BranchingIndex(){
		double B=0;
		for(unsigned v=0; v < V.size(); v ++){
			B += ((V[v].e[0].size()) + (V[v].e[1].size()));
		}
		B = B / V.size();
		return B;

	}

	// Returns number of branch points in thenetwork
	unsigned int BranchP(){
		unsigned int B=0;
		unsigned int c;
		for(unsigned v=0; v < V.size(); v ++){
			c = ((V[v].e[0].size()) + (V[v].e[1].size()));
			if (c > 2){
			B += 1;}
		}		
		return B;

	}

	// Returns number of end points (tips) in thenetwork
	unsigned int EndP(){
		unsigned int B=0;
		unsigned int c;
		for(unsigned v=0; v < V.size(); v ++){
			c = ((V[v].e[0].size()) + (V[v].e[1].size()));
			if (c == 1){
			B += 1;}
		}		
		return B;

	}

	//// Returns a dictionary with the key as the vertex
	//std::map<std::vector<vertex>,unsigned int> DegreeDict(){
	//	std::map<std::vector<vertex>,unsigned int> dd;
	//	unsigned int c = 0;
	//	for(unsigned v=0; v < V.size(); v ++){
	//		c = ((V[v].e[0].size()) + (V[v].e[1].size()));
	//		dd[V[v]] = c;
	//	}
	//	return dd;
	//}

	//// Return number of branching stems
	//unsigned int Stems(){
	//	unsigned int s = 0;
	//	std::map<std::vector<vertex>,unsigned int> dd;
	//	dd = DegreeDict();
	//	//for(unsigned v=0; v < V.size(); v ++){
	//	//	V[v].e[0].
	//	return s;
	//}

	//Calculate Metrics---------------------------------------------------
	// Returns an average of fiber/edge lengths in the network
	double Lengths(){
		stim::vec<T> L;
		double sumLength = 0;
		for(unsigned e = 0; e < E.size(); e++){				//for each edge in the network
			L.push_back(E[e].length());						//append the edge length
			sumLength = sumLength + E[e].length();
		}
		double avg = sumLength / E.size();
		return avg;
	}


	// Returns an average of tortuosities in the network
	double Tortuosities(){
		stim::vec<T> t;
		stim::vec<T> id1, id2;                        // starting and ending vertices of the edge
		double distance;double tortuosity;double sumTortuosity = 0;
		for(unsigned e = 0; e < E.size(); e++){				//for each edge in the network
			id1 = E[e][0];									//get the edge starting point
			id2 = E[e][E[e].size() - 1];					//get the edge ending point
			distance = (id1 - id2).len();                   //displacement between the starting and ending points
			if(distance > 0){
				tortuosity = E[e].length()/	distance	;		// tortuoisty = edge length / edge displacement
			}
			else{
				tortuosity = 0;}
			t.push_back(tortuosity);
			sumTortuosity += tortuosity;
		}
		double avg = sumTortuosity / E.size();
		return avg;
	}

	// Returns average contraction of the network
	double Contractions(){
	stim::vec<T> t;
	stim::vec<T> id1, id2;                        // starting and ending vertices of the edge
	double distance;double contraction;double sumContraction = 0;
	for(unsigned e = 0; e < E.size(); e++){				//for each edge in the network
		id1 = E[e][0];									//get the edge starting point
		id2 = E[e][E[e].size() - 1];					//get the edge ending point
		distance = (id1 - id2).len();                   //displacement between the starting and ending points
		contraction = distance / E[e].length();		// tortuoisty = edge length / edge displacement
		t.push_back(contraction);
		sumContraction += contraction;
	}
	double avg = sumContraction / E.size();
	return avg;
	}

	// returns average fractal dimension of the branches of the network
	double FractalDimensions(){
	stim::vec<T> t;
	stim::vec<T> id1, id2;                        // starting and ending vertices of the edge
	double distance;double fract;double sumFractDim = 0;
	for(unsigned e = 0; e < E.size(); e++){				//for each edge in the network
		id1 = E[e][0];									//get the edge starting point
		id2 = E[e][E[e].size() - 1];					//get the edge ending point
		distance = (id1 - id2).len();                   //displacement between the starting and ending points
		fract = std::log(distance) / std::log(E[e].length());		// tortuoisty = edge length / edge displacement
		t.push_back(sumFractDim);
		sumFractDim += fract;
	}
	double avg = sumFractDim / E.size();
	return avg;
	}

	//returns a cylinder represented a given fiber (based on edge index)
	stim::cylinder<T> get_cylinder(unsigned e){
		return E[e];									//return the specified edge (casting it to a fiber)
	}

	//load a network from an OBJ file
	void load_obj(std::string filename){

		stim::obj<T> O;									//create an OBJ object
		O.load(filename);								//load the OBJ file as an object

		std::vector<unsigned> id2vert;							//this list stores the OBJ vertex ID associated with each network vertex

		unsigned i[2];									//temporary, IDs associated with the first and last points in an OBJ line

		//for each line in the OBJ object
		for(unsigned int l = 1; l <= O.numL(); l++){

			std::vector< stim::vec<T> > c;						//allocate an array of points for the vessel centerline
			O.getLine(l, c);							//get the fiber centerline

			stim::centerline<T> c3(c.size());
			for(size_t j = 0; j < c.size(); j++)
				c3[j] = c[j];
			c3.update();

	//		edge new_edge = c3;		///This is dangerous.
			edge new_edge(c3);
					
			//create an edge from the given centerline
			unsigned int I = new_edge.size();					//calculate the number of points on the centerline

			//get the first and last vertex IDs for the line
			std::vector< unsigned > id;						//create an array to store the centerline point IDs
			O.getLinei(l, id);							//get the list of point IDs for the line
			i[0] = id.front();							//get the OBJ ID for the first element of the line
			i[1] = id.back();							//get the OBJ ID for the last element of the line

			std::vector<unsigned>::iterator it;					//create an iterator for searching the id2vert array
			unsigned it_idx;							//create an integer for the id2vert entry index

			//find out if the nodes for this fiber have already been created
			it = find(id2vert.begin(), id2vert.end(), i[0]);	//look for the first node
			if(it == id2vert.end()){							//if i[0] hasn't already been used
				vertex new_vertex = new_edge[0];				//create a new vertex, assign it a position
				new_vertex.e[0].push_back(E.size());				//add the current edge as outgoing
				new_edge.v[0] = V.size();					//add the new edge to the edge
				V.push_back(new_vertex);					//add the new vertex to the vertex list
				id2vert.push_back(i[0]);					//add the ID to the ID->vertex conversion list
			}
			else{									//if the vertex already exists
				it_idx = std::distance(id2vert.begin(), it);
				V[it_idx].e[0].push_back(E.size());				//add the current edge as outgoing
				new_edge.v[0] = it_idx;
			}

			it = find(id2vert.begin(), id2vert.end(), i[1]);			//look for the second ID
			if(it == id2vert.end()){						//if i[1] hasn't already been used
				vertex new_vertex = new_edge[I-1];				//create a new vertex, assign it a position
				new_vertex.e[1].push_back(E.size());				//add the current edge as incoming
				new_edge.v[1] = V.size();                                  	//add the new vertex to the edge
				V.push_back(new_vertex);					//add the new vertex to the vertex list
				id2vert.push_back(i[1]);					//add the ID to the ID->vertex conversion list
			}
			else{									//if the vertex already exists
				it_idx = std::distance(id2vert.begin(), it);
				V[it_idx].e[1].push_back(E.size());				//add the current edge as incoming
				new_edge.v[1] = it_idx;
			}

			E.push_back(new_edge);							//push the edge to the list

		}
	}

	//load a network from an SWC file
	void load_swc(std::string filename) {
		stim::swc<T> S;										// create swc variable
		S.load(filename);									// load the node information
		S.create_tree();									// link those node according to their linking relationships as a tree

		//NT.push_back(S.node[0].type);						// set the neuronal_type value to the first vertex in the network
		std::vector<unsigned> id2vert;						// this list stores the SWC vertex ID associated with each network vertex
		unsigned i[2];										// temporary, IDs associated with the first and last points
		for (unsigned int l = 1; l < S.numL(); l++) {		// for every node starts from second one
			//NT.push_back(S.node[l].type);
			stim::centerline<T> c3(2);						// every fiber contains two vertices
			int p_idx = S.node[l].parent_idx - 1;			// get the parent node loop idx of current node
			c3[0] = S.node[p_idx].point;					// store the begin vertex
			c3[1] = S.node[l].point;						// store the end vertex
		
			c3.update();									// update the L information
			
			stim::cylinder<T> C3(c3);						// create a new cylinder in order to copy the origin radius information
			// upadate the R information
			std::vector<T> radius;
			radius.push_back(S.node[p_idx].radius);
			radius.push_back(S.node[l].radius);
			C3.copy_r(radius);

			edge new_edge(C3);								// contains two points								
			
			//get the first and last vertex IDs for the line
			i[0] = S.node[p_idx].idx;						//get the SWC ID for the start point
			i[1] = S.node[l].idx;							//get the SWC ID for the end point

			std::vector<unsigned>::iterator it;				//create an iterator for searching the id2vert array
			unsigned it_idx;								//create an integer for the id2vert entry index

			//find out if the nodes for this fiber have already been created
			it = find(id2vert.begin(), id2vert.end(), i[0]);	//look for the first node
			if (it == id2vert.end()) {							//if i[0] hasn't already been used
				vertex new_vertex = new_edge[0];				//create a new vertex, assign it a position
				new_vertex.e[0].push_back(E.size());			//add the current edge as outgoing
				new_edge.v[0] = V.size();						//add the new edge to the edge
				V.push_back(new_vertex);						//add the new vertex to the vertex list
				id2vert.push_back(i[0]);						//add the ID to the ID->vertex conversion list
			}
			else {									//if the vertex already exists
				it_idx = std::distance(id2vert.begin(), it);
				V[it_idx].e[0].push_back(E.size());				//add the current edge as outgoing
				new_edge.v[0] = it_idx;
			}

			it = find(id2vert.begin(), id2vert.end(), i[1]);	//look for the second ID
			if (it == id2vert.end()) {							//if i[1] hasn't already been used
				vertex new_vertex = new_edge[1];			//create a new vertex, assign it a position
				new_vertex.e[1].push_back(E.size());			//add the current edge as incoming
				new_edge.v[1] = V.size();                       //add the new vertex to the edge
				V.push_back(new_vertex);						//add the new vertex to the vertex list
				id2vert.push_back(i[1]);						//add the ID to the ID->vertex conversion list
			}
			else {									//if the vertex already exists
				it_idx = std::distance(id2vert.begin(), it);
				V[it_idx].e[1].push_back(E.size());				//add the current edge as incoming
				new_edge.v[1] = it_idx;
			}

			E.push_back(new_edge);								//push the edge to the list
		}
	}

	/// Output the network as a string
	std::string str(){

		std::stringstream ss;
		ss<<"Nodes ("<<V.size()<<")--------"<<std::endl;
		for(unsigned int v = 0; v < V.size(); v++){
			ss<<"\t"<<v<<V[v].str()<<std::endl;
		}

		ss<<"Edges ("<<E.size()<<")--------"<<std::endl;
		for(unsigned e = 0; e < E.size(); e++){
			ss<<"\t"<<e<<E[e].str()<<std::endl;
		}

		return ss.str();
	}

	/// This function resamples all fibers in a network given a desired minimum spacing
	/// @param spacing is the minimum distance between two points on the network
	stim::network<T> resample(T spacing){
		stim::network<T> n;								//create a new network that will be an exact copy, with resampled fibers
		n.V = V;									//copy all vertices
		//n.NT = NT;										//copy all the neuronal type information
		n.E.resize(edges());								//allocate space for the edge list

		//copy all fibers, resampling them in the process
		for(unsigned e = 0; e < edges(); e++){						//for each edge in the edge list
			n.E[e] = E[e].resample(spacing);					//resample the edge and copy it to the new network
		}

		return n;							              	//return the resampled network
	}

	// host gaussian function
	__host__ float gaussianFunction(float x, float std = 25){ return exp(-x/(2*std*std));}  // std default value is 25

	/// Calculate the total number of points on all edges.
	unsigned total_points(){
		unsigned n = 0;
		for(unsigned e = 0; e < E.size(); e++)
			n += E[e].size();
		return n;
	}

	//Copy the point cloud representing the centerline for the network into an array
	void centerline_cloud(T* dst) {
		size_t p;										//stores the current edge point
		size_t P;										//stores the number of points in an edge
		size_t i = 0;									//index into the output array of points
		for (size_t e = 0; e < E.size(); e++) {			//for each edge in the network
			P = E[e].size();							//get the number of points in this edge
			for (p = 0; p < P; p++) {
				dst[i * 3 + 0] = E[e][p][0];		
				dst[i * 3 + 1] = E[e][p][1];
				dst[i * 3 + 2] = E[e][p][2];
				i++;
			}
		}
	}

    // convert vec3 to array
	void stim2array(float *a, stim::vec3<T> b){
		a[0] = b[0];
		a[1] = b[1];
		a[2] = b[2];
	}

	// convert vec3 to array in bunch
	void edge2array(T* a, edge b){
		size_t n = b.size();
		for(size_t i = 0; i < n; i++){
			a[i * 3 + 0] = b[i][0];
			a[i * 3 + 1] = b[i][1];
			a[i * 3 + 2] = b[i][2];	 
		}
	}

	/// Calculate the average magnitude across the entire network.
	/// @param m is the magnitude value to use. The default is 0 (usually radius).
	T average(unsigned m = 0){

		T M, L;										//allocate space for the total magnitude and length
		M = L = 0;									//initialize both the initial magnitude and length to zero
		for(unsigned e = 0; e < E.size(); e++){						//for each edge in the network
			M += E[e].integrate();							//get the integrated magnitude
			L += E[e].length();							//get the edge length
		}

		return M / L;									//return the average magnitude
	}

	/// This function compares two networks and returns the percentage of the current network that is missing from A.
	/// @param A is the network to compare to - the field is generated for A
	/// @param sigma is the user-defined tolerance value - smaller values provide a stricter comparison
	stim::network<T> compare(stim::network<T> A, float sigma, int device = -1){

		stim::network<T> R;										//generate a network storing the result of the comparison
		R = (*this);											//initialize the result with the current network

		T *c;						                 			// centerline (array of double pointers) - points on kdtree must be double
		size_t n_data = A.total_points();          				// set the number of points
		c = (T*) malloc(sizeof(T) * n_data * 3);				// allocate an array to store all points in the data set				

		unsigned t = 0;
		for(unsigned e = 0; e < A.E.size(); e++){				//for each edge in the network
			for(unsigned p = 0; p < A.E[e].size(); p++){		//for each point in the edge
				for(unsigned d = 0; d < 3; d++){				//for each coordinate

					c[t * 3 + d] = A.E[e][p][d];				//copy the point into the array c
				}
				t++;
			}
		}

		//generate a KD-tree for network A
		size_t MaxTreeLevels = 3;								// max tree level
		
#ifdef __CUDACC__
		cudaSetDevice(device);
		stim::kdtree<T, 3> kdt;								// initialize a pointer to a kd tree

		kdt.create(c, n_data, MaxTreeLevels);				// build a KD tree

		for(unsigned e = 0; e < R.E.size(); e++){					//for each edge in A
			//R.E[e].add_mag(0);							//add a new magnitude for the metric
			//size_t errormag_id = R.E[e].nmags() - 1;		//get the id for the new magnitude
			
			size_t n = R.E[e].size();						// the number of points in current edge
			T* queryPt = new T[3 * n];
			T* m1 = new T[n];
			T* dists = new T[n];
			size_t* nnIdx = new size_t[n];

			T* d_dists;										
			T* d_m1;										
			cudaMalloc((void**)&d_dists, n * sizeof(T));
			cudaMalloc((void**)&d_m1, n * sizeof(T));

			edge2array(queryPt, R.E[e]);
			kdt.search(queryPt, n, nnIdx, dists);		

			cudaMemcpy(d_dists, dists, n * sizeof(T), cudaMemcpyHostToDevice);					// copy dists from host to device

			// configuration parameters
			size_t threads = (1024>n)?n:1024;
			size_t blocks = n/threads + (n%threads)?1:0;

			d_metric<<<blocks, threads>>>(d_m1, n, d_dists, sigma);					//calculate the metric value based on the distance

			cudaMemcpy(m1, d_m1, n * sizeof(T), cudaMemcpyDeviceToHost);

			for(unsigned p = 0; p < n; p++){
				R.E[e].set_r(p, m1[p]);
			}

			//d_set_mag<<<blocks, threads>>>(R.E[e].M, errormag_id, n, m1);
		}

#else
		stim::kdtree<T, 3> kdt;
		kdt.create(c, n_data, MaxTreeLevels);
	
		for(unsigned e = 0; e < R.E.size(); e++){			//for each edge in A
			R.E[e].add_mag(0);								//add a new magnitude for the metric
			size_t errormag_id = R.E[e].nmags() - 1;

			size_t n = R.E[e].size();						// the number of points in current edge
			T* query = new T[3 * n];
			T* m1 = new T[n];
			T* dists = new T[n];
			size* nnIdx = new size_t[n];

			edge2array(queryPt, R.E[e]);

			kdt.cpu_search(queryPt, n, nnIdx, dists);			//find the distance between A and the current network

			for(unsigned p = 0; p < R.E[e].size(); p++){
				m1[p] = 1.0f - gaussianFunction((T)dists[p], sigma);		//calculate the metric value based on the distance
				R.E[e].set_mag(errormag_id, p, m1[p]);					//set the error for the second point in the segment
			}
		}
#endif
		return R;		//return the resulting network
	}

	/// This function compares two networks and split the current one according to the nearest neighbor of each point in each edge
	/// @param A is the network to split
	/// @param B is the corresponding mapping network
	/// @param sigma is the user-defined tolerance value - smaller values provide a stricter comparison
	/// @param device is the device that user want to use
	void split(stim::network<T> A, stim::network<T> B, float sigma, int device, float threshold){

		T *c;						                 	
		size_t n_data = B.total_points();          				
		c = (T*) malloc(sizeof(T) * n_data * 3); 				

		size_t NB = B.E.size();								// the number of edges in B
		unsigned t = 0;
		for(unsigned e = 0; e < NB; e++){					// for every edge in B			
			for(unsigned p = 0; p < B.E[e].size(); p++){	// for every points in B.E[e]
				for(unsigned d = 0; d < 3; d++){				

					c[t * 3 + d] = B.E[e][p][d];			// convert to array
				}
				t++;
			}
		}
		size_t MaxTreeLevels = 3;							// max tree level

#ifdef __CUDACC__
		cudaSetDevice(device);
		stim::kdtree<T, 3> kdt;								// initialize a pointer to a kd tree
	
		//compare each point in the current network to the field produced by A
		kdt.create(c, n_data, MaxTreeLevels);				// build a KD tree

		std::vector<std::vector<unsigned>> relation;		// the relationship between GT and T corresponding to NN
		relation.resize(A.E.size());										

		for(unsigned e = 0; e < A.E.size(); e++){			//for each edge in A
			//A.E[e].add_mag(0);								//add a new magnitude for the metric
			//size_t errormag_id = A.E[e].nmags() - 1;
			
			size_t n = A.E[e].size();						// the number of edges in A

			T* queryPt = new T[3 * n];							// set of all the points in current edge
			T* m1 = new T[n];								// array of metrics for every point in current edge
			T* dists = new T[n];							// store the distances for every point in current edge
			size_t* nnIdx = new size_t[n];					// store the indices for every point in current edge
			
			// define pointers in device
			T* d_dists;														
			T* d_m1;
			size_t* d_nnIdx;

			// allocate memory for defined pointers
			cudaMalloc((void**)&d_dists, n * sizeof(T));
			cudaMalloc((void**)&d_m1, n * sizeof(T));
			cudaMalloc((void**)&d_nnIdx, n * sizeof(size_t));

			edge2array(queryPt, A.E[e]);						// convert edge to array
			kdt.search(queryPt, n, nnIdx, dists);				// search the tree to find the NN for every point in current edge

			cudaMemcpy(d_dists, dists, n * sizeof(T), cudaMemcpyHostToDevice);					// copy dists from host to device
			cudaMemcpy(d_nnIdx, nnIdx, n * sizeof(size_t), cudaMemcpyHostToDevice);				// copy Idx from host to device

			// configuration parameters
			size_t threads = (1024>n)?n:1024;													// test to see whether the number of point in current edge is more than 1024
			size_t blocks = n/threads + (n%threads)?1:0;

			d_metric<<<blocks, threads>>>(d_m1, n, d_dists, sigma);								// calculate the metrics in parallel

			cudaMemcpy(m1, d_m1, n * sizeof(T), cudaMemcpyDeviceToHost);

			for(unsigned p = 0; p < n; p++){
				A.E[e].set_r(p, m1[p]);											// set the error(radius) value to every point in current edge
			}

			relation[e].resize(n);																// resize every edge relation size

			unsigned* d_relation;
			cudaMalloc((void**)&d_relation, n * sizeof(unsigned));								// allocate memory

			std::vector<size_t> edge_point_num(NB);												// %th element is the number of points that %th edge has
			for(unsigned ee = 0; ee < NB; ee++)
				edge_point_num[ee] = B.E[ee].size();

			size_t* d_edge_point_num;
			cudaMalloc((void**)&d_edge_point_num, NB * sizeof(size_t));
			cudaMemcpy(d_edge_point_num, &edge_point_num[0], NB * sizeof(size_t), cudaMemcpyHostToDevice);

			d_findedge<<<blocks, threads>>>(d_nnIdx, n, d_relation, d_edge_point_num, NB);			// find the edge corresponding to the array index in parallel

			cudaMemcpy(&relation[e][0], d_relation, n * sizeof(unsigned), cudaMemcpyDeviceToHost);	//copy relationship from device to host
		}
#else
		stim::kdtree<T, 3> kdt;
		kdt.create(c, n_data, MaxTreeLevels);
	
		std::vector<std::vector<unsigned>> relation;		// the mapping relationship between two networks
		relation.resize(A.E.size());										
		for(unsigned i = 0; i < A.E.size(); i++)
			relation[i].resize(A.E[i].size());

		std::vector<size_t> edge_point_num(NB);				//%th element is the number of points that %th edge has
		for(unsigned ee = 0; ee < NB; ee++)
			edge_point_num[ee] = B.E[ee].size();

		for(unsigned e = 0; e < A.E.size(); e++){			//for each edge in A
			A.E[e].add_mag(0);								//add a new magnitude for the metric
			size_t errormag_id = A.E[e].nmags() - 1;

			size_t n = A.E[e].size();						// the number of edges in A

			T* queryPt = new T[3 * n];
			T* m1 = new T[n];
			T* dists = new T[n];							//store the distances
			size_t* nnIdx = new size_t[n];					//store the indices
			
			edge2array(queryPt, A.E[e]);
			kdt.search(queryPt, n, nnIdx, dists);		

			for(unsigned p = 0; p < A.E[e].size(); p++){
				m1[p] = 1.0f - gaussianFunction((T)dists[p], sigma);	//calculate the metric value based on the distance
				A.E[e].set_mag(errormag_id, p, m1[p]);					//set the error for the second point in the segment
				
				unsigned id = 0;																	//mapping edge's idx
				size_t num = 0;																		//total number of points before #th edge
				for(unsigned i = 0; i < NB; i++){
					num += B.E[i].size();
					if(nnIdx[p] < num){																//find the edge it belongs to
						relation[e][p] = id;
						break;
					}
					id++;																			//current edge won't be the one, move to next edge
				}
			}
		}
#endif
		E = A.E;
		V = A.V;

		unsigned int id = 0;									// split value								
		for(unsigned e = 0; e < E.size(); e++){					// for every edge
			for(unsigned p = 0; p < E[e].size() - 1; p++){		// for every point in each edge
				int t = E[e].length() / sigma * 2;
				if (t <= 20)
					threshold_fac = E[e].size();
				else
					threshold_fac = (E[e].length() / sigma * 2)/10;
				if(relation[e][p] != relation[e][p + 1]){		// find the nearest edge changing point
					id = p + 1;									// if there is no change in NN
					if(id < threshold_fac || (E[e].size() - id) < threshold_fac)				
						id = E[e].size() - 1;																			// extreme situation is not acceptable
					else
						break;
				}
				if(p == E[e].size() - 2)																// if there is no splitting index, set the id to the last point index of current edge
					id = E[e].size() - 1;
			}
			//unsigned errormag_id = E[e].nmags() - 1;
			T G = 0;																					// test to see whether it has its nearest neighbor
			for(unsigned i = 0; i < E[e].size(); i++)
				G += E[e].r(i);																			// won't split special edges
			if(G / E[e].size() > threshold)															// should based on the color map
				id = E[e].size() - 1;																	// set split idx to outgoing direction vertex

			std::vector<edge> tmpe;
			tmpe.resize(2);
			tmpe = E[e].split(id);
			vertex tmpv = stim::vec3<T>(-1, -1, 0);														// store the split point as vertex
			if(tmpe.size() == 2){
				relation.resize(relation.size() + 1);
				for(unsigned d = id; d < E[e].size(); d++)
					relation[relation.size() - 1].push_back(relation[e][d]);
				tmpe[0].v[0] = E[e].v[0];																// begining vertex of first half edge -> original begining vertex
				tmpe[1].v[1] = E[e].v[1];																// ending vertex of second half edge -> original ending vertex
				tmpv = E[e][id];
				V.push_back(tmpv);
				tmpe[0].v[1] = (unsigned)V.size() - 1;													// ending vertex of first half edge -> new vertex
				tmpe[1].v[0] = (unsigned)V.size() - 1;													// begining vertex of second half edge -> new vertex
				edge tmp(E[e]);
				E[e] = tmpe[0];																			// replace original edge by first half edge
				E.push_back(tmpe[1]);																	// push second half edge to the last
				V[V.size() - 1].e[1].push_back(e);														// push first half edge to the incoming of new vertex
				V[V.size() - 1].e[0].push_back((unsigned)E.size() - 1);									// push second half edge to the outgoing of new vertex
				for(unsigned i = 0; i < V[tmp.v[1]].e[1].size(); i++)									// find the incoming edge of original ending vertex
					if(V[tmp.v[1]].e[1][i] == e)
						V[tmp.v[1]].e[1][i] = (unsigned)E.size() - 1;									// set to new edge
			}
		}
	}

	/// This function compares two splitted networks and yields a mapping relationship between them according to NN
	/// @param B is the network that the current network is going to map to
	/// @param C is the mapping relationship: C[e1] = _e1 means e1 edge in current network is mapping the _e1 edge in B
	/// @param device is the device that user want to use
	void mapping(stim::network<T> B, std::vector<unsigned> &C, int device, float threshold){
		stim::network<T> A;								//generate a network storing the result of the comparison
		A = (*this);

		size_t n = A.E.size();							// the number of edges in A
		size_t NB = B.E.size();							// the number of edges in B

		C.resize(A.E.size());	

		T *c;						                 	// centerline (array of double pointers) - points on kdtree must be double
		size_t n_data = B.total_points();          		// set the number of points
		c = (T*) malloc(sizeof(T) * n_data * 3); 				

		unsigned t = 0;
		for(unsigned e = 0; e < NB; e++){					// for each edge in the network
			for(unsigned p = 0; p < B.E[e].size(); p++){	// for each point in the edge
				for(unsigned d = 0; d < 3; d++){			// for each coordinate

					c[t * 3 + d] = B.E[e][p][d];
				}
				t++;
			}
		}

		//generate a KD-tree for network A
		//float metric = 0.0;                               		// initialize metric to be returned after comparing the network
		size_t MaxTreeLevels = 3;									// max tree level
		
#ifdef __CUDACC__
		cudaSetDevice(device);
		stim::kdtree<T, 3> kdt;								// initialize a pointer to a kd tree
	
		kdt.create(c, n_data, MaxTreeLevels);				// build a KD tree

		for(unsigned e = 0; e < n; e++){					//for each edge in A
			//size_t errormag_id = A.E[e].nmags() - 1;		//get the id for the new magnitude
			
			//pre-judge to get rid of impossibly mapping edges
			T M = 0;
			for(unsigned p = 0; p < A.E[e].size(); p++)
				M += A.E[e].r(p);
			M = M / A.E[e].size();
			if(M > threshold)
				C[e] = (unsigned)-1;						//set the nearest edge of impossibly mapping edges to maximum of unsigned
			else{
				T* queryPt = new T[3];
				T* dists = new T[1];
				size_t* nnIdx = new size_t[1];

				stim2array(queryPt, A.E[e][A.E[e].size()/2]);
				kdt.search(queryPt, 1, nnIdx, dists);
				
				unsigned id = 0;							//mapping edge's idx
				size_t num = 0;								//total number of points before #th edge
				for(unsigned i = 0; i < NB; i++){
					num += B.E[i].size();
					if(nnIdx[0] < num){
						C[e] = id;
						break;
					}
					id++;
				}
			}
		}
#else
		stim::kdtree<T, 3> kdt;
		kdt.create(c, n_data, MaxTreeLevels);
		T *dists = new T[1];								// near neighbor distances
		size_t *nnIdx = new size_t[1];						// near neighbor indices // allocate near neigh indices

		stim::vec3<T> p0, p1;
		T* queryPt = new T[3];

		for(unsigned e = 0; e < R.E.size(); e++){			// for each edge in A
			T M;											// the sum of metrics of current edge
			unsigned errormag_id = R.E[e].nmags() - 1;
			for(unsigned p = 0; p < R.E[e].size(); p++)
				M += A.E[e].m(p, errormag_id);
			M = M / A.E[e].size();
			if(M > threshold)								
				C[e] = (unsigned)-1;
			else{											// if it should have corresponding edge in B, then...
				p1 = R.E[e][R.E[e].size()/2];							
				stim2array(queryPt, p1);
				kdt.cpu_search(queryPt, 1, nnIdx, dists);	// search the tree		
				
				unsigned id = 0;							//mapping edge's idx
				size_t num = 0;								//total number of points before #th edge
				for(unsigned i = 0; i < NB; i++){
					num += B.E[i].size();
					if(nnIdx[0] < num){
						C[e] = id;
						break;
					}
					id++;
				}
			}
		}
#endif
	}

	/// Returns the number of magnitude values stored in each edge. This should be uniform across the network.
	//unsigned nmags(){
	//	return E[0].nmags();
	//}
	// split a string in text by the character sep
	stim::vec<T> split(std::string &text, char sep) 
	{
		stim::vec<T> tokens;
		std::size_t start = 0, end = 0;
		while ((end = text.find(sep, start)) != std::string::npos) {
		tokens.push_back(atof(text.substr(start, end - start).c_str()));
		start = end + 1;
		}
		tokens.push_back(atof(text.substr(start).c_str()));
		return tokens;
	}
	// load a network in text file to a network class
	void load_txt(std::string filename)
	{
		std::vector <std::string> file_contents;
		std::ifstream file(filename.c_str());
		std::string line;
		std::vector<unsigned> id2vert;	//this list stores the vertex ID associated with each network vertex
		//for each line in the text file, store them as strings in file_contents
		while (std::getline(file, line))
		{
			std::stringstream ss(line);
			file_contents.push_back(ss.str());
		}
		unsigned int numEdges = atoi(file_contents[0].c_str()); //number of edges in the network
		unsigned int I = atoi(file_contents[1].c_str()) ;		//calculate the number of points3d on the first edge
		unsigned int count = 1; unsigned int k = 2; // count is global counter through the file contents, k is for the vertices on the edges
		// for each edge in the network.
		for (unsigned int i = 0; i < numEdges; i ++ )
		{
			// pre allocate a position vector p with number of points3d on the edge p
			std::vector< stim::vec<T> > p(0, I);
			// for each point on the nth edge
		  for (unsigned int j = k; j < I + k; j++)
		  {
			 // split the points3d of floats with separator space and form a float3 position vector out of them
			  p.push_back(split(file_contents[j], ' '));
		  }
			count += p.size() + 1; // increment count to point at the next edge in the network
			I = atoi(file_contents[count].c_str()); // read in the points3d at the next edge and convert it to an integer
			k = count + 1;
			edge new_edge = p; // create an edge with a vector of points3d  on the edge
			E.push_back(new_edge); // push the edge into the network
		}
		unsigned int numVertices = atoi(file_contents[count].c_str()); // this line in the text file gives the number of distinct vertices
		count = count + 1; // this line of text file gives the first verrtex
		// push each vertex into V
		for (unsigned int i = 0; i < numVertices; i ++)
		{
			vertex new_vertex = split(file_contents[count], ' ');
			V.push_back(new_vertex);
			count += atoi(file_contents[count + 1].c_str()) + 2; // Skip number of edge ids + 2 to point to the next vertex
		}
	} // end load_txt function

	// strTxt returns a string of edges
	std::string
	strTxt(std::vector< stim::vec<T> > p)
	{
		std::stringstream ss;
		std::stringstream oss;
		for(unsigned int i = 0; i < p.size(); i++){
			ss.str(std::string());
			for(unsigned int d = 0; d < 3; d++){
				ss<<p[i][d];
			}
			ss << "\n";
		}
		return ss.str();
	}
	// removes specified character from string
	void removeCharsFromString(std::string &str, char* charsToRemove ) {
	   for ( unsigned int i = 0; i < strlen(charsToRemove); ++i ) {
		  str.erase( remove(str.begin(), str.end(), charsToRemove[i]), str.end() );
	   }
	}
	//exports network to txt file
	void
	to_txt(std::string filename)
	{
		std::ofstream ofs(filename.c_str(), std::ofstream::out | std::ofstream::app);
		//int num;
		ofs << (E.size()).str() << "\n";
		for(unsigned int i = 0; i < E.size(); i++)
		{
			 std::string str;
			 ofs << (E[i].size()).str() << "\n";
			 str = E[i].strTxt();
             ofs << str << "\n";
		} 
		for(int i = 0; i < V.size(); i++)
		{
			std::string str;
			str = V[i].str();
			char temp[4] = "[],";
			removeCharsFromString(str, temp);
			ofs << str << "\n";
		}
		ofs.close();
	}
};		//end stim::network class
};		//end stim namespace
#endif